unit Win32_3964R;

interface

uses Classes, Windows;

{ V1.3

  This implementation of the 3964R-protocol is based on a thread class, that
  performs the communication in the background while a the main program thread
  can continue with its normal operations.

  The class listens to the servial device all the time and is always ready
  to accept incoming telegrams. A send fifo queue is supported for up to
  100 telegrams waiting to be sended out.

  This implementation doesn't cover multi-block transmissions and receivs. Only
  a single block per telegram up to a length of MaxNetDataPerBlock (default=128,
  can be changed) can be handled.

  You can create multiple instances of this class for different devices. Each
  will empoy its own thread.

  I don't see any need for a developer to derive descendants from this class.
  However I made all the protected methods virtual to maximize flexibility.

  The main thread will get notifications on received and sended telegrams and
  errors over (optional) event procedures.

  The public methods SendTelegram and TransactTelegram are called by the main
  thread to instruct the class to send data. These methods are not reentrant
  safe. So don't call these methods (of one instance) more than once at the
  same time from different threads.


  Here are some short instructions about howto use this class:

  Method Create
    Parameters:
    - See comments at the method definition below.
    Note: The paramater ADoSyncEvents control if the event methods (AOn3964XXXX)
          are called using the Synchronize mechanism of the vcl TThread class
          or not. If you choose false (no sync), the event methods will be called
          in the thread's context w/o being synchronized to anything. If you
          choose true (do sync), you can't make use of the wait options
          (AWaitQueueSpace and AWaitSended) of the SendTelegram method and you
          can't use TransactTelegram at all.
          Be sure not to call methods of the calling TCom3964Thread instance from
          within you event handler in either case.

  Method TerminateAndFree
    Call this method instead of .Free to destroy a class instance. This terminates
    the running thread properly.

  Method SendTelegram
    Parameters:
    - ASend gives the data to be sended out.
    - AWaitQueueSpace controls wether the method will wait until there is a free
          slot in the send queue in case the queue is full.
    - AWaitSended controls wether the method will wait until the given data has been
          sended out (not recommended).
    - ASendId is a pointer to a DWORD variable that will receive an id value for
          the telegram queued to send. This can be nil if you don't need this id.
          This id will be referenced in the On3964Received event method to report
          success or failure of sending this telegram.

    Return value:
    True  ASend was placed into the send queue.
          Also, with parameter AWaitSended=true, the result true means that
          delivery of AData to the communication partner was successful.
    False No slot was free in the send queue and AWaitQueueSpace was false.
          If AWaitQueueSpace AWaitSended are both given true, the return value
          will be false if sending out the data failed.

  Method TransactTelegram
    This method sends a telegram out (with AWaitSended option) and then waits
    approx. up to the time specified in TransactionAnswerTimeout for an anwser
    from the communication partner to come in.

    Parameters:
    - ASend: Same as for SendTelegram
    - ARecv will hold the first telegram received after sending out ASend
    - AWaitQueueSpace: Same as for SendTelegram

    Return value:
    True  ASend was placed into the send queue and sended out with success.
          After sending out ASend, there came in a telegram (ARecv) from the
          partner within aprox. the time specified in TransactionAnswerTimeout.
    False No slot was free in the send queue and AWaitQueueSpace was false.
          Or sending out AData resulted in an (protocol) error.
          Or after sending out AData there came no aswer within the timeout.

  Field FHint (public):
    Can be used by developers for any purpose. Not used by the class itself.

  On3964Received event:
    Notification of received telegram. AData holds the net data that came in.
    Telegrams that are received within a transaction are not reported this way,
    but returned by the ARecv of the method TransactTelegram

  On3964Sent event:
    A telegram that was placed into the send queue using SendTelegram or
    TransactTelegram was sended out (ASuccess=true) or sending resulted in a
    protocol error (ASuccess=false).
    AId identifies the telegram with the same value that was given to the
    caller of SendTelegram via the ASendId parameter.
    AId is zero if SendTelegram was called with ASendId=nil or if sending
    this telegram was issued by TransactTelegram.

  On3964Error event:
    Called by code running in the thread context to notify the host thread about
    an error. The AMsg string contains text describing the error. Can for example
    be a protocol error of the 3964R procedure (for example block check error).
    In such cases the thread will continue to run (ATerminates=false). Other
    reasons might for example be exceptions when calling winapi functions.
    Exceptions are reported with ATerminates=true. The thread then dies.
    Exceptions that occur in the host thread context (e.g. in Create, SendTelegram,
    TransactTelegram) are not reported via this event handler.

  Executing property:
    This read-only property is true if the thread is running. The value of this
    property is set to true in .Execute and never becomes false again.

  DeviceName property:
    This read-only property gives you the name of the device in use (e.g. 'COM1'
    for AComNo=1).
}


{Global constants. Don't change after creating a TCom3964Thread.}
const
  MaxNetDataPerBlock: Integer = 128; //note: multiblock is not supported
  QVZ: Integer = 2000; //time (ms) within a reaction on sended data is expected to come in
  MaxSendRetries: Integer = 4; //max count of tries to send a telegram
  TransactionAnswerTimeout = 30; //max time to wait for transaction answer in 1/10 sec.

type
  TCom3964Thread = class;

  TOn3964Received = procedure(Sender: TCom3964Thread; AData: string) of object;
  TOn3964Sent = procedure(Sender: TCom3964Thread; AId: Integer; ASuccess: Boolean) of object;
  TOn3964Error = procedure(Sender: TCom3964Thread; AMsg: string; ATerminates: Boolean) of object;

  TCom3964Thread = class(TThread)
  private
    FComm: THandle;
    FOvlThread,
    FOvlMain: TOverlapped;
    FExecuting: Boolean;
    FSendRecvEvent: THandle;
    FDevStr: string;
    FHighPrio: Boolean;
    FSendTGs: array of string;
    FSendTG_Ids: array of DWORD;
    FCurrentSendStr: string;
    FSendSTXIssued: Boolean;
    FSendTgTryCnt,
    FSendStackEntries: Integer;
    FOn3964Received: TOn3964Received;
    FOn3964Sent: TOn3964Sent;
    FOn3964Error: TOn3964Error;
    FDoSyncEvents: Boolean;
    FIdle: Boolean;
    FProtectReentrance,
    FLastSendId: Integer;
    FTransactWaitOnRecv: Boolean;
    FNotifyReceived: string;
    FNotifySendId: DWORD;
    FNotifySendSuccess: Boolean;
    FNotifyErrMsg: string;
    FNotifyIsException: Boolean;
    procedure GetNextSendTG;
    procedure ShiftSendTgs;
    procedure WriteCommFile(const s: string; AOvl: POverlapped);
    procedure ReadCommFile(var ABuf; ABuflen: DWORD; var ABytesRead: DWORD; AOvl: POverlapped);
    function InternalSendTelegram(const ASend: string;
                                  AWaitQueueSpace, AWaitSended: Boolean;
                                  ASendId: PDWORD;
                                  ATransaction: Boolean): Boolean;
  protected
    FAccessSendRecvTG: TRTLCriticalSection;
    FAccessIdle: TRTLCriticalSection;
    procedure Notify3964Received; virtual;
    procedure Notify3964Sent; virtual;
    procedure Notify3964Error; virtual;
    procedure ReceiveDone(s: string); virtual;
    procedure SendDone(ASuccess: Boolean); virtual;
    procedure ThreadError(AErrMsg: string; IsException: Boolean = false); virtual;
  public
    FHint: Integer; //free to use by developers for any purpose
    constructor Create(AComNo, //e.g. 1 for COM1
                       ABaudrate, //use windows consts CBR_9600
                       AByteSize, //usually 8
                       AStopBits, //use windows consts like ONESTOPBIT
                       AParity: Integer; //use windows consts like NOPARITY
                       ARtsCts, //true=user rts/cts lines for hardware protocol
                       AHighPrio: Boolean; //true=high 3964 priority
                       ASendStackEntries: Integer; //min=1, max=100
                       AOn3964Received: TOn3964Received; //optional receive notification
                       AOn3964Sent: TOn3964Sent; //optional sent notification
                       AOn3964Error: TOn3964Error; //optional error notification
                       ADoSyncEvents: Boolean); //false=call AOn3964XXXX in thread context
    destructor Destroy; override;
    procedure Execute; override;
    procedure TerminateAndFree;
    function SendTelegram(const ASend: string;
                          AWaitQueueSpace, AWaitSended: Boolean;
                          ASendId: PDWORD): Boolean;
    function TransactTelegram(const ASend: string; var ARecv: string; AWaitQueueSpace: Boolean): Boolean;
    property Executing: boolean read FExecuting;
    property DeviceName: string read FDevStr;
  end;

implementation

uses Sysutils, Math;

const
  STX = #2;
  ETX = #3;
  ETB = #23;
  DLE = #16;
  NAK = #21;

resourcestring
  SUnexpectedCharRecvd = 'Unexpected char received';
  SErrorReceivingData = 'Protocol error receiving data';
  STransactTelegramIllegalReentrance = 'Illegal Reentrance in TCom3964Thread.TransactTelegram';
  STransactTelegramWaitNotAllowed = 'TCom3964Thread.TransactTelegram not allowed when '+
    'DoSyncEvents=true (to prevent deadlocks)';
  SSendTelegramWaitNotAllowed = 'TCom3964Thread.SendTelegram: AWaitQueueSpace and AWaitSended '+
    'not allowed when DoSyncEvents=true (to prevent deadlocks)';
  SInternalSendTelegramIllegalReentrance = 'Illegal Reentrance in TCom3964Thread.InternalSendTelegram';
  SSendTelegramIllegalLenInternal = 'Illegal length of ASend Data passed to '+
    'TCom3964Thread.InternalSendTelegram';

constructor TCom3964Thread.Create(AComNo,
                                  ABaudrate,
                                  AByteSize,
                                  AStopBits,
                                  AParity: Integer;
                                  ARtsCts,
                                  AHighPrio: Boolean;
                                  ASendStackEntries: Integer;
                                  AOn3964Received: TOn3964Received;
                                  AOn3964Sent: TOn3964Sent;
                                  AOn3964Error: TOn3964Error;
                                  ADoSyncEvents: Boolean);
var
  DCB: TDCB;
  CommTimeOuts: TCommTimeOuts;
begin
  InitializeCriticalSection(FAccessSendRecvTG);
  InitializeCriticalSection(FAccessIdle);
  FDevStr := 'COM' + IntToStr(AComNo);
  FHighPrio := AHighPrio;
  FOn3964Received := AOn3964Received;
  FOn3964Sent := AOn3964Sent;
  FOn3964Error := AOn3964Error;
  FDoSyncEvents := ADoSyncEvents and (Assigned(FOn3964Received) or
                                      Assigned(FOn3964Sent) or
                                      Assigned(FOn3964Error));

  FComm := CreateFile(PChar(FDevStr),
                      GENERIC_READ or GENERIC_WRITE,
                      0, nil, OPEN_EXISTING,
                      FILE_ATTRIBUTE_SYSTEM or FILE_FLAG_OVERLAPPED, 0);
  if FComm = INVALID_HANDLE_VALUE then RaiseLastWin32Error;

  FOvlThread.hEvent := CreateEvent(nil, true, false, nil);
  FOvlMain.hEvent := CreateEvent(nil, true, false, nil);

  ZeroMemory(@DCB, SizeOf(DCB));
  with DCB do
  begin
    DCBlength := SizeOf(DCB);
    Baudrate := ABaudrate;
    ByteSize := AByteSize;
    Parity := AParity;
    StopBits := AStopBits;
    Flags := $0001; //binary
    if Aparity <> NOPARITY then Flags := Flags or $0002; //enable-parity-checking
    if ARtsCts then Flags := Flags or $1004; //cts-out-control and turn-rts-on
  end;
  with CommTimeOuts do
  begin
    ReadIntervalTimeout := MAXDWORD;
    ReadTotalTimeoutMultiplier := MAXDWORD;
    ReadTotalTimeoutConstant := QVZ;
    WriteTotalTimeoutMultiplier := 2;
    WriteTotalTimeoutConstant := 100;
  end;
  Win32Check(SetCommState(FComm, DCB));
  Win32Check(SetCommTimeOuts(FComm, CommTimeOuts));
  FSendStackEntries := Max(Min(ASendStackEntries, 100), 1);
  SetLength(FSendTGs, FSendStackEntries);
  SetLength(FSendTG_Ids, FSendStackEntries);
  FSendRecvEvent := CreateEvent(nil, true, false, nil);
  if FSendRecvEvent = 0 then RaiseLastWin32Error;
  inherited Create(false);
end;

destructor TCom3964Thread.Destroy;
begin
  if (FComm <> INVALID_HANDLE_VALUE) and (FComm <> 0) then CloseHandle(FComm);
  if FOvlThread.hEvent <> 0 then CloseHandle(FOvlThread.hEvent);
  if FOvlMain.hEvent <> 0 then CloseHandle(FOvlMain.hEvent);
  if FSendRecvEvent <> 0 then CloseHandle(FSendRecvEvent);
  DeleteCriticalSection(FAccessIdle);
  DeleteCriticalSection(FAccessSendRecvTG);
  inherited Destroy;
end;

procedure TCom3964Thread.Execute;

          procedure SetIdle(Value: Boolean);
          begin
            EnterCriticalSection(FAccessIdle);
            FIdle := Value;
            LeaveCriticalSection(FAccessIdle);
          end;

type
  Tcheckstate = (csNone, csHaveDLE, csHaveETX, csHaveBCC);
var
  ChrRecvd: boolean;
  c: char;
  s, s2: string;
  bcc: byte;
  ok, err: boolean;
  i, j: Integer;
  BytesRead: DWORD;
  checkstate: Tcheckstate;
begin
  FExecuting := true;
  try
    repeat
      ChrRecvd := false;
      GetNextSendTG;
      if FCurrentSendStr = '' then
        repeat
          SetIdle(true);
          try
            ReadCommFile(c, 1, BytesRead, @FOvlThread);
          finally
            SetIdle(false);
          end;
          if Terminated then Exit;
          ChrRecvd := BytesRead = 1;
          GetNextSendTG;
        until ChrRecvd or (FCurrentSendStr <> '');

      //send a telegram
      if (FCurrentSendStr <> '') and ((not ChrRecvd) or FSendStxIssued) then
      begin
        if not FSendStxIssued then WriteCommFile(STX, @FOvlThread);
        FSendStxIssued := false;
        Inc(FSendTgTryCnt);

        if not ChrRecvd then
        begin
          ReadCommFile(c, 1, BytesRead, @FOvlThread);
          ChrRecvd := BytesRead = 1
        end;

        if not ChrRecvd or not (c in [DLE, STX]) then
        begin
          if ChrRecvd and (c <> NAK) then
          begin
            Win32Check(PurgeComm(FComm, PURGE_RXCLEAR));
            WriteCommFile(NAK, @FOvlThread);
            ThreadError(SUnexpectedCharRecvd);
          end;
          if FSendTgTryCnt = MaxSendRetries then SendDone(false);
          continue;
        end;

        if (c = DLE) or (FHighPrio) then
        begin
          bcc := 0;
          SetLength(s, Length(FCurrentSendStr)*2+3); //maxlen=all dle*2 + 3(dle,etx,bcc)
          j := 1;
          for i := 1 to Length(FCurrentSendStr) do
          begin
            c := FCurrentSendStr[i];
            s[j] := c;
            bcc := bcc xor byte(c);
            Inc(j);
            if c = DLE then
            begin
              s[j] := c;
              Inc(j);
              bcc := bcc xor byte(c);
            end;
          end;
          s[j] := DLE;
          s[j+1] := ETX;
          s[j+2] := char(bcc xor byte(DLE) xor byte(ETX));
          SetLength(s, j+2);

          WriteCommFile(s, @FOvlThread);

          ReadCommFile(c, 1, BytesRead, @FOvlThread);
          ChrRecvd := BytesRead = 1;

          if ChrRecvd and (c = DLE) then
            SendDone(true)
          else
          begin
            if ChrRecvd and (c <> NAK) then
            begin
              Win32Check(PurgeComm(FComm, PURGE_RXCLEAR));
              WriteCommFile(NAK, @FOvlThread);
              ThreadError(SUnexpectedCharRecvd);
            end;
            if FSendTgTryCnt = MaxSendRetries then SendDone(false);
          end;

          continue;
        end
        else
          Dec(FSendTgTryCnt); //conflicting stx+stx and lowprio doesn't count
      end;

      if not ChrRecvd then continue;

      //receive a telegram
      if c = STX then
      begin
        WriteCommFile(DLE, @FOvlThread);

        SetLength(s2, MaxNetDataPerBlock * 2 + 3); //max=all dle*2 + 3(dle,etx,bcc)
        s := '';
        ok := false;
        err := false;
        bcc := 0;
        checkstate := csNone;
        repeat
          ReadCommFile(s2[1], Length(s2), BytesRead, @FOvlThread);
          if BytesRead > 0 then
          begin
            for i := 1 to BytesRead do
            begin
              c := s2[i];
              case checkstate of
                csNone: begin
                  bcc := bcc xor byte(c);
                  if c = DLE then checkstate := csHaveDLE
                             else s := s + c;
                end;
                csHaveDLE: begin
                  bcc := bcc xor byte(c);
                  case c of
                    ETX: checkstate := csHaveETX;
                    DLE: begin
                      checkstate := csNone;
                      s := s + c;
                    end;
                  else
                    err := true;
                    break;
                  end;
                end;
                csHaveETX: begin
                  if c = char(bcc) then
                  begin
                    checkstate := csHaveBCC;
                    ok := true;
                  end
                  else
                  begin
                    err := true;
                    break;
                  end;
                end;
                csHaveBCC: begin
                  err := true; //received char after BCC illegal
                  break;
                end;
              end;
            end;
          end;
        until ok or err or (BytesRead = 0);
        ok := ok and not err and (s <> '');

        if ok then
        begin
          WriteCommFile(DLE, @FOvlThread);
          ReceiveDone(s);
        end
        else
        begin
          Win32Check(PurgeComm(FComm, PURGE_RXCLEAR));
          WriteCommFile(NAK, @FOvlThread);
          ThreadError(SErrorReceivingData);
        end;
      end;
    until Terminated;
  except
    on E: Exception do
    begin
      if not Terminated then
        if Assigned(FOn3964Error) then ThreadError(E.Message, true)
                                  else raise;
    end;
  end;
end;

procedure TCom3964Thread.TerminateAndFree;
begin
  Terminate;
  PurgeComm(FComm, PURGE_TXABORT or PURGE_RXABORT); //kick thread out of ReadFile
  WaitFor;
  Free;
end;

procedure TCom3964Thread.WriteCommFile(const s: string; AOvl: POverlapped);
var dw: DWORD;
begin
  if not WriteFile(FComm, s[1], Length(s), dw, AOvl) then
    if GetLastError <> ERROR_IO_PENDING then RaiseLastWin32Error
    else
      Win32Check(GetOverlappedResult(FComm, AOvl^, dw, true));
end;

procedure TCom3964Thread.ReadCommFile(var ABuf; ABuflen: DWORD; var ABytesRead: DWORD; AOvl: POverlapped);
begin
  if not ReadFile(FComm, ABuf, ABuflen, ABytesRead, AOvl) then
    if GetLastError <> ERROR_IO_PENDING then RaiseLastWin32Error
    else
      Win32Check(GetOverlappedResult(FComm, AOvl^, ABytesRead, true));
end;

procedure TCom3964Thread.Notify3964Received;
begin
  FOn3964Received(Self, FNotifyReceived);
end;

procedure TCom3964Thread.Notify3964Sent;
begin
  FOn3964Sent(Self, FNotifySendId, FNotifySendSuccess);
end;

procedure TCom3964Thread.Notify3964Error;
begin
  FOn3964Error(Self, FNotifyErrMsg, FNotifyIsException);
end;

procedure TCom3964Thread.ReceiveDone(s: string);
begin
  EnterCriticalSection(FAccessSendRecvTG);
  try
    FNotifyReceived := s; //needed in Notify3964Received and in TransactTelegram
  finally
    LeaveCriticalSection(FAccessSendRecvTG);
  end;

  if Assigned(FOn3964Received) and (not Terminated) and (not FTransactWaitOnRecv) then
    if FDoSyncEvents then Synchronize(Notify3964Received) else Notify3964Received;

  Win32Check(PulseEvent(FSendRecvEvent));
end;

procedure TCom3964Thread.SendDone(ASuccess: Boolean);
begin
  FSendTgTryCnt := 0;
  FNotifySendSuccess := ASuccess; //needed in Notify3964Sent and in InternalSendTelegram
  if Assigned(FOn3964Sent) and not Terminated then
    if FDoSyncEvents then Synchronize(Notify3964Sent) else Notify3964Sent;
  ShiftSendTgs;
  Win32Check(PulseEvent(FSendRecvEvent));
end;

procedure TCom3964Thread.ThreadError(AErrMsg: string; IsException: Boolean = false);
begin
  if Assigned(FOn3964Error) and not Terminated then
  begin
    FNotifyErrMsg := AErrMsg;
    FNotifyIsException := IsException;
    if FDoSyncEvents then Synchronize(Notify3964Error) else Notify3964Error;
  end;
end;

procedure TCom3964Thread.GetNextSendTG;
begin
  EnterCriticalSection(FAccessSendRecvTG);
  try
    FCurrentSendStr := FSendTGs[0];
    FNotifySendId := FSendTG_Ids[0];
  finally
    LeaveCriticalSection(FAccessSendRecvTG);
  end;
end;

procedure TCom3964Thread.ShiftSendTgs;
var
  i: Integer;
  s: string;
begin
  EnterCriticalSection(FAccessSendRecvTG);
  try
    FSendTGs[0] := '';
    FSendTG_Ids[0] := 0;
    if (FSendStackEntries > 1) and (FSendTGs[1] <> '') then
      for i := 1 to FSendStackEntries-1 do
      begin
        s := FSendTGs[i];
        FSendTGs[i-1] := s;
        FSendTG_Ids[i-1] := FSendTG_Ids[i];
        if s = '' then break;
        if i = FSendStackEntries-1 then
        begin
          FSendTGs[i] := '';
          FSendTG_Ids[i] := 0;
        end;
      end;

    FNotifyReceived := ''; //clear this buffer now, a new transaction mugh now start
  finally
    LeaveCriticalSection(FAccessSendRecvTG);
  end;
end;

function TCom3964Thread.InternalSendTelegram(const ASend: string;
                                             AWaitQueueSpace, AWaitSended: Boolean;
                                             ASendId: PDWORD;
                                             ATransaction: Boolean): Boolean;
var
  i: Integer;
  QueueEmpty: Boolean;
begin
  Result := false;
  if Assigned(ASendId) then ASendId^ := 0;
   
  i := InterlockedIncrement(FProtectReentrance);
  try
    if ((i = 2) and not ATransaction) or (i > 2) then
      raise Exception.Create(SInternalSendTelegramIllegalReentrance);

    if (ASend = '') or (Length(ASend) > MaxNetDataPerBlock) then
      raise Exception.Create(SSendTelegramIllegalLenInternal);

    repeat
      EnterCriticalSection(FAccessSendRecvTG);
      try
        for i := 0 to FSendStackEntries-1 do
          if FSendTGs[i] = '' then
          begin
            FSendTGs[i] := ASend;
            if ATransaction then
              FSendTG_Ids[i] := 0
            else
            begin
              Inc(FLastSendId);
              if FLastSendId = 0 then Inc(FLastSendId);
              FSendTG_Ids[i] := FLastSendId;
            end;
            if Assigned(ASendId) then ASendId^ := FSendTG_Ids[i];
            Result := true;
            if (i = 0) and not FSendSTXIssued then
            begin
              EnterCriticalSection(FAccessIdle);
              try
                if FIdle then
                begin
                  WriteCommFile(STX, @FOvlMain);
                  FSendSTXIssued := true;
                end;
              finally
                LeaveCriticalSection(FAccessIdle);
              end;
            end;
            break;
          end;
      finally
        LeaveCriticalSection(FAccessSendRecvTG);
      end;

      if (not Result) and AWaitQueueSpace then
      begin
        if WaitForSingleObject(FSendRecvEvent, 100) = WAIT_FAILED then
          RaiseLastWin32Error;
      end;
    until Result or (not AWaitQueueSpace);

    if Result and AWaitSended then
    begin
      repeat
        if WaitForSingleObject(FSendRecvEvent, 100) = WAIT_FAILED then
          RaiseLastWin32Error;
        EnterCriticalSection(FAccessSendRecvTG);
        try
          QueueEmpty := FSendTGs[0] = '';
        finally
          LeaveCriticalSection(FAccessSendRecvTG);
        end;
      until QueueEmpty;
      Result := FNotifySendSuccess;
    end;
  finally
    InterlockedDecrement(FProtectReentrance);
  end;
end;

function TCom3964Thread.SendTelegram(const ASend: string;
                                     AWaitQueueSpace,
                                     AWaitSended: Boolean;
                                     ASendId: PDWORD): Boolean;
begin
  if FDoSyncEvents and (AWaitQueueSpace or AWaitSended) then
    raise Exception.Create(SSendTelegramWaitNotAllowed);

  Result := InternalSendTelegram(ASend, AWaitQueueSpace, AWaitSended, ASendId, false);
end;

function TCom3964Thread.TransactTelegram(const ASend: string; var ARecv: string;
                                         AWaitQueueSpace: Boolean): Boolean;
var
  i: Integer;
  WaitsOnRecv: Integer;
begin
  Result := false; //TACW
  
  if FDoSyncEvents then
    raise Exception.Create(STransactTelegramWaitNotAllowed);

  i := InterlockedIncrement(FProtectReentrance);
  try
    if (i > 1) then
      raise Exception.Create(STransactTelegramIllegalReentrance);

    ARecv := '';
    Result := InternalSendTelegram(ASend, AWaitQueueSpace, true, nil, true);
    if Result then
    begin
      FTransactWaitOnRecv := true;
      try
        WaitsOnRecv := 0;
        repeat
          if WaitForSingleObject(FSendRecvEvent, 100) = WAIT_FAILED then
            RaiseLastWin32Error;
          EnterCriticalSection(FAccessSendRecvTG);
          try
            Inc(WaitsOnRecv);
            ARecv := FNotifyReceived;
          finally
            LeaveCriticalSection(FAccessSendRecvTG);
          end;
          Result := ARecv <> '';
        until Result or (WaitsOnRecv = TransactionAnswerTimeout);
      finally
        FTransactWaitOnRecv := false;
      end;
    end;
  finally
    InterlockedDecrement(FProtectReentrance);
  end;
end;

end.
